# Name: config.py
# Description: defines configurations for the various components of audio extraction and processing


class AudioConfig:
    # The format to store audio in
    AUDIO_FORMAT = 'mp3'
    # Prefix to save audio features to
    FEATURE_DESTINATION = '/features/'
    # Checkpoint frequency in number of tracks processed
    CHECKPOINT_FREQUENCY = 10
    # Minimum required clip length for prediction
    MIN_CLIP_LENGTH = 29


class DisplayConfig:
    # What cmap to use when saving visual features
    # Refer to https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
    CMAP = "Greys"
    # Defines the size of the figures created by display
    FIGSIZE_WIDTH = 10
    FIGSIZE_HEIGHT = 10


class FeatureExtractorConfig:
    # The Librosa features supported by the CLI
    SUPPORTED_FEATURES = ['chroma_stft', 'rms', 'spec_cent', 'spec_bw', 'spec_rolloff', 'zcr', 'mfcc']
    NUMBER_OF_MFCC_COLS = 20
    # How to aggregate the features
    FEATURE_AGGREGATION = ['mean', 'min', 'max', 'std']
    N_FFT = 2048
    HOP_LENGTH = 1024



class YoutubeExtractionConfig:
    # URL to make request to for audio results
    SEARCH_URL = 'https://www.youtube.com/results'
    # Query parameters to filter requests to search URL by
    SEARCH_QUERY_PARAMS = ['search_query']
    # URL to make request to for actual audio data
    CONTENT_URL = 'https://www.youtube.com/watch'
    # Query parameters to filter requests to content URL by
    CONTENT_QUERY_PARAMS = ['v']
    # Post-processor key to use
    POST_PROCESSOR_KEY = 'FFmpegExtractAudio'
    # Post-processor audio quality
    POST_PROCESSOR_QUALITY = '192'
    # Preferred audio quality name
    PREFERRED_AUDIO_QUALITY = 'bestaudio/best'


class SongExtractorConfig:
    # A list of supported sources by the package
    SUPPORTED_SOURCES = ['youtube']
    # Where to save audio files to
    DESTINATION_FILEPATH = '/Documents/audio-clips/'



import requests


class Request(object):

    def __init__(self, base_url):
        """ Instantiates unfiltered request object with just the given URL that will be used to make requests to
        different endpoints
        :param string base_url: the starting URL of the endpoint to use without any filter query parameters
        """
        self.filtered = False
        self.url = base_url

    def filter_by(self, filter_param, filter_value):
        """ Adds request query parameters in the form of ?filter_param=filter_value to the
        existing url
        :param string filter_param: the name of the filter query parameter
        :param string filter_value: the corresponding value of the query parameter
        """
        # The first time this method is called the results will not be filtered
        # The filter should be appended with ?
        if not self.filtered:
            url_format = '{0}?{1}={2}'
            self.filtered = True
        # The subsequent times this method is called, each filter should be appended with &
        else:
            url_format = '{0}&{1}={2}'
        self.url = url_format.format(self.url, filter_param, filter_value)

    def get(self, url=None):
        """ Submits a GET request to an API endpoint
        :param string url: an optional URL to use instead of the one created through constructor
        :returns a response object from the requests library
        """
        if not url:
            url = self.url
        response = requests.get(url)
        # Raise an exception if request is invalid
        response.raise_for_status()
        # Otherwise return the response
        return response



import logging


from genreml.model.acquisition.youtube import YouTubeDownloader
from genreml.model.acquisition.config import SongExtractorConfig
from genreml.model.processing.config import AudioConfig
from genreml.model.utils import file_handling


class SongExtractor(object):

    def __init__(self, source=SongExtractorConfig.SUPPORTED_SOURCES[0], config=SongExtractorConfig):
        """ Instantiates SongExtractor object to interact with downloaders and download audio files from given search
        queries
        :param string source: a supported source to extract audio files from
        :param model.processing.config.SongExtractorConfig config: a SongExtractor config object
        """
        self.config = config

        # The given source must be supported
        if source not in config.SUPPORTED_SOURCES:
            raise ValueError("song source {0} is not one of {1}".format(source, config.SUPPORTED_SOURCES))
        self.source = source
        # Get the associated downloader
        self.downloader = self.get_downloader(source)

    @staticmethod
    def get_downloader(source):
        """ Factory method for different downloaders used by SongExtractor to extract music from the web
        :param string source: the name of the downloader to use (refer to SongExtractorConfig)
        """
        if source == 'youtube':
            return YouTubeDownloader()

    def extract(self, song_name, artist, file_path=None):
        """ Extracts data for the given song name and artist using the active downloader
        :param string song_name: the name of the song to download
        :param string artist: the name of the song's artist
        :param file_path: an optional path to save the song file to (must be an absolute path including the song file)
        """
        # If no full path given, construct one from config
        if not file_path:
            file_path = self.config.DESTINATION_FILEPATH
            full_path = file_handling.get_directory_path(file_path)
            file_handling.create_directory(full_path)
            file_name = '{0}_{1}_clip.{2}'.format(song_name, artist, AudioConfig.AUDIO_FORMAT)
            file_path = "{0}{1}".format(full_path, file_name)
        logging.info("Changing working directory to {0}".format(file_path))
        file_handling.change_directory_to(file_path)
        search_query = '{0}+{1}'.format(song_name, artist)
        self.downloader.download(file_path, search_values=[search_query])



import os

from genreml.model.processing import audio
from genreml.model.utils import file_handling


def process_audio_file(root_directory, location: str) -> tuple:
    """ Given a location of an uploaded file, this will run feature extraction and return a tuple containing lists
    of features + images to be displayed to the user
    :param root_directory - the root directory the app is running from
    :param location - the location of a file to extract audio features from
    :returns tuple containing the lists of features extracted
    """
    # Visual features will crash Flask app if the following is not configured
    import matplotlib
    matplotlib.use('Agg')
    # Run feature extraction
    audio_files = audio.AudioFiles()
    audio_format = file_handling.get_filetype(location).replace(".", "")
    audio_files.extract_features(location, cmap=None, figure_height=3, figure_width=5, audio_format=audio_format)
    directory = os.path.join(root_directory, "static")
    visual_feature_paths = []
    for idx, visual_feature in enumerate(audio_files.visual_features[0]):
        # Create unique name for each file generated from visual features
        image_name = file_handling.get_unique_file_name(str(idx))
        full_path = "{0}/{1}.png".format(directory, image_name)
        # Save each image
        visual_feature.savefig(full_path)
        visual_feature_paths.append(full_path)
    return audio_files.features, visual_feature_paths




from genreml.model.acquisition import youtube
from genreml.model.utils import file_handling


def download_link(youtube_url: str, directory_path: str, audio_format: str = "wav") -> str:
    """ Downloads a track from the given youtube_url into the given directory_path

    :param youtube_url - the URL to a youtube video containing the audio file to download
    :param directory_path - the path to a local directory in which to save the downloaded file
    :param audio_format - the format of the audio to download
    :returns the full path where the audio was saved to
    """
    assert audio_format in {"wav", "mp3"}
    downloader = youtube.YouTubeDownloader()
    unique_video_id = file_handling.get_unique_file_name(youtube_url)
    full_path = "{0}/{1}.{2}".format(directory_path, unique_video_id, audio_format)
    downloader.download(full_path, url=youtube_url)
    return full_path



import numpy as np
import tensorflow as tf

from genreml.model.cnn import config
from genreml.model.model import base_model, input


class CnnInput(input.ModelInput):

    def __init__(self, spectrograms: list, features: list, name: str = "ConvolutionalNeuralNetworkInput") -> None:
        super().__init__(name=name, spectrograms=spectrograms, features=features)


class CnnModel(base_model.Model):

    def __init__(self, name: str = "ConvolutionalNeuralNetwork", model_config=config.CnnModelConfig):
        super().__init__(name)
        self.config = model_config

    def train(self) -> None:
        pass

    def _preprocess_spectrogram(self, image: list) -> np.array:
        image = np.array(image).reshape(self.config.IMG_HEIGHT, self.config.IMG_WIDTH, 1)
        return image

    def predict(self, input_data: input.ModelInput) -> np.array:
        if not self.model:
            raise AttributeError("Model {0} is not trained".format(self.name))
        elif "spectrograms" not in input_data or "features" not in input_data:
            raise AttributeError(
                "Both spectrograms and raw features need to be provided to model {0}".format(self.name))
        spectrograms = self._preprocess_spectrogram(input_data["spectrograms"])
        features = np.array(input_data["features"])
        prediction = self.model.predict([np.array([features]), np.array(spectrograms)])
        return prediction

    @classmethod
    def from_h5_file(cls, h5_filepath: str):
        cls_instance = cls()
        cls_instance.model = tf.keras.models.load_model(h5_filepath)
        return cls_instance



import pandas as pd
import pkg_resources


def get_resource_csv(filename: str) -> pd.DataFrame:
    """ Retrieves a CSV file from the model resources packaged up with genreml

    :param filename - the name of the file to retrieve
    :returns a data frame constructed from the CSV file
    """
    if filename[-4:] != ".csv":
        raise ValueError("Filename {0} does not correspond to a CSV".format(filename))
    directory = pkg_resources.resource_filename('genreml', 'model_resources/')
    return pd.read_csv(directory + filename)



import os
import sys

from genreml.model.utils.file_handling import get_filetype, get_filename
from flask import request, render_template
from werkzeug.utils import secure_filename

from frontend import app
from frontend.processing import feature_extraction, genre_prediction, downloader


def populate_visual_features(root_path, filepath):
    features, visual_features = feature_extraction.process_audio_file(root_path, filepath)
    visual_feature_relative_paths = ["/static/{0}".format(get_filename(feature)) for feature in visual_features]
    visual_feature_relative_paths = [tuple(visual_feature_relative_paths[i:i + 2])
                                     for i in range(0, len(visual_feature_relative_paths), 2)]
    os.remove(filepath)
    return visual_feature_relative_paths


@app.route('/upload', methods=['POST'])
def upload_file():
    """ Route for uploading audio files to the site """
    if request.method == 'POST':
        # Get the file information from the upload
        f = request.files['file']
        filename = secure_filename(f.filename)
        extension = get_filetype(filename).replace(".", "")
        # Process the file only if extension is allowed
        if extension in app.config['ALLOWED_EXTENSIONS']:
            filepath = os.path.join(app.config['FILE_UPLOAD_LOCATION'], filename)
            f.save(filepath)
            visual_feature_relative_paths = populate_visual_features(app.root_path, filepath)
            return render_template('results.html', visual_features=visual_feature_relative_paths)
        return render_template('index.html')


@app.route('/download_youtube', methods=['POST'])
def process_youtube_video():
    """ Route for downloading a Youtube video and performing feature extraction + prediction """
    if request.method == 'POST':
        # Get the Youtube link from the request
        youtube_url = request.form['text']
        # Only process results if valid youtube video link
        if "youtube" in youtube_url and youtube_url[:8] == "https://" and "v=" in youtube_url:
            filepath = downloader.download_link(youtube_url="", directory_path=app.config['FILE_UPLOAD_LOCATION'])
            visual_feature_relative_paths = populate_visual_features(app.root_path, filepath)
            return render_template('results.html', visual_features=visual_feature_relative_paths)
    return render_template('index.html')


@app.route('/')
def index():
    """ Server route for the app's landing page """
    return render_template('index.html')


if __name__ == "__main__":
    environment = "production"
    if len(sys.argv) > 1:
        environment = sys.argv[1]
    if environment == "production":
        from gevent.pywsgi import WSGIServer
        app.debug = True
        http_server = WSGIServer(('', 8000), app)
        http_server.serve_forever()
    elif environment == "development":
        app.run(debug=True, host='127.0.0.1', port=8000)



# Name: extraction.py
# Description: defines functionality to download audio files from the web

import youtube_dl
import re

from genreml.model.acquisition.config import YoutubeExtractionConfig
from genreml.model.acquisition import request


class YouTubeDownloader(object):

    def __init__(self, config=YoutubeExtractionConfig):
        """ Instantiates Youtube downloader with a config object as defined in model.processing.config by default
        :param model.processing.config.YoutubeExtractionConfig config: a Youtube extraction config object
        """
        self.config = config
        config_attributes = config.__dict__
        # Config object must have URLs for retrieving the video IDs from search queries and downloading the actual vids
        assert 'SEARCH_URL' in config_attributes and 'CONTENT_URL' in config_attributes

    @staticmethod
    def video_id_from_url(video_url: str) -> str:
        """ Extracts the video ID from a Youtube video URL

        :param video_url - the URL to a Youtube video (example - www.youtube.com/watch?v=FDMq9ie0ih0)
        :returns the video ID
        """
        if "v=" not in video_url:
            raise ValueError("Video URL {0} does not contain a valid video ID parameter".format(video_url))
        video_id = video_url.strip("v=")[1]
        return video_id

    def _get_video_id_to_download(self, search_values):
        """ Retrieves a video ID to download based on song and artist search keywords
        :param list search_values: a list of search keywords corresponding to each search parameter in Youtube search
        :returns a single video ID that best matches the given search values
        """
        # There must be as many search values as expected from the config specification
        if len(search_values) != len(self.config.SEARCH_QUERY_PARAMS):
            raise ValueError("The given search filters {0} do not match what's expected: {1}".format(
                search_values, self.config.SEARCH_QUERY_PARAMS))

        # Construct search request to extract YouTube video results
        search_request = request.Request(self.config.SEARCH_URL)
        for search_param, search_value in zip(self.config.SEARCH_QUERY_PARAMS, search_values):
            search_request.filter_by(search_param, search_value)
        # Get data from search query
        search_response = search_request.get().text
        video_ids = re.findall(r"watch\?v=(\S{11})", search_response)
        # TODO Currently just takes the first match but we may want to add additional parsing logic to pick the best
        result_video_id = video_ids[0]
        return result_video_id

    def download(self, destination_filepath: str, search_values: list = None, url: str = None):
        """ Download a video matching the given search_values to the given destination file path
        :param destination_filepath: the full file path including the file name to download
        :param list search_values: a list of search keywords corresponding to each search parameter in Youtube search
        :param url: the content URL to download directly
        """
        content_url = url
        if search_values:
            # Get the best matching video ID from the given search values
            video_id_to_download = self._get_video_id_to_download(search_values)
            # Create a request object and filter by the video ID extracted above
            content_request = request.Request(self.config.CONTENT_URL)
            video_filter_param = self.config.CONTENT_QUERY_PARAMS[0]
            content_request.filter_by(video_filter_param, video_id_to_download)
            # Get the full Youtube URL pointing to the best matching video to download
            content_url = content_request.url
        # Download the video to the destination_filepath
        with youtube_dl.YoutubeDL({
            'format': self.config.PREFERRED_AUDIO_QUALITY,
            'outtmpl': destination_filepath
        }) as ydl:
            ydl.download([content_url])





class ModelInput(object):

    def __init__(self, name: str, **kwargs) -> None:
        self.name = name
        for input_name, input_value in kwargs:
            self.__dict__[input_name] = input_value

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def __repr__(self) -> str:
        return "Model input {0}".format(self.name)

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(self, item: str) -> any:
        return self.__dict__.get(item)



from genreml.model.utils import multithreading


class MockReceiver(object):

    def __init__(self):
        self.records = []
        self.random_strings = []

    def insert(self, records: list, some_random_string: str):
        self.records += records
        self.random_strings += [some_random_string for _ in records]


def test_run_function_in_threads():
    """ Tests utils.multithreading.run_function_in_threads function """
    receiver = MockReceiver()
    data = [x for x in range(18)]
    length_of_data = len(data)
    original_data = data.copy()
    number_of_threads = 10
    multithreading.run_function_in_thread(number_of_threads, receiver.insert, data, "some_random_string")
    assert len(receiver.records) == length_of_data and len(receiver.random_strings) == length_of_data
    assert receiver.records == original_data \
           and receiver.random_strings == ["some_random_string" for _ in range(length_of_data)]



import numpy as np

from abc import ABC, abstractmethod

from genreml.model.model import input


class Model(ABC):

    def __init__(self, name: str) -> None:
        self.name = name
        self.model = None

    @abstractmethod
    def train(self):
        pass

    @abstractmethod
    def predict(self, input_data: input.ModelInput) -> np.array:
        pass
